﻿--[[
	LootMaster, Master looter stuff
]]--

local debug	= false

LootMasterML	    = LibStub("AceAddon-3.0"):NewAddon("LootMasterML", "AceConsole-3.0", "AceComm-3.0", "AceEvent-3.0")
local LootMaster    = LibStub("AceAddon-3.0"):GetAddon("EPGPLootMaster")

local addon         = LootMasterML		-- Local instance of the addon

function LootMasterML:Debug( message )
    if not LootMaster or not LootMaster.debug then return end;
    self:Print("debug: " .. message)
end

function LootMasterML:OnInitialize()

	self.lootTable = {}			-- a table to store loot info

	-- Event Register
    -- Trap event when ML rightclicks master loot
	self:RegisterEvent("OPEN_MASTER_LOOT_LIST")  
    
    -- Trap even when an items get looted
    self:RegisterEvent("CHAT_MSG_LOOT");
    
    -- Trap events when entering and leaving combat
    self:RegisterEvent("PLAYER_REGEN_DISABLED",     "EnterCombat");
    self:RegisterEvent("PLAYER_REGEN_ENABLED",      "LeaveCombat");
    
    self:RegisterEvent("GUILD_ROSTER_UPDATE",       "CacheGuildInfo")   

	-- Register communications
	self:RegisterComm("EPGPLootMasterML", 		    "CommandReceived")
        
    -- Create table for the guildinfo cache.
    self.guildInfo = {}
    
    -- Disable 'automatic loot tracking' popup in EPGP - Let EPGPLootmaster handle all the GP stuff.
    if EPGP and CheckForGuildInfo and IsInGuild() then
        CheckForGuildInfo(EPGP);        
        if EPGP.db then
            EPGP.db.profile.auto_loot = false
        end
    end
    
end

function LootMasterML:GetVersionString()
    local lm = LibStub("AceAddon-3.0"):GetAddon("EPGPLootMaster")
    return (lm.GetVersionString(lm) or 'unknown version')
end

function LootMasterML:SendCommand(command, message, target)
	if not target then
		return self:Print("Could not send command, no target specified")
	end;
    if target=='RAID' then
        self:SendCommMessage("EPGPLootMasterC", format("%s:%s", tostring(command), tostring(message)), "RAID", nil, "ALERT")
    elseif target=='PARTY' then
        self:SendCommMessage("EPGPLootMasterC", format("%s:%s", tostring(command), tostring(message)), "PARTY", nil, "ALERT")
    else
        self:SendCommMessage("EPGPLootMasterC", format("%s:%s", tostring(command), tostring(message)), "WHISPER", target, "ALERT")
    end
end

function LootMasterML:ParseMonitorMessage( message )
    local tpl = {}
    for v in string.gmatch(message, '(.-)(%^%*)') do
        if v=='nil' then v=nil end;
        tinsert(tpl, v)
    end
    for _,_,v in string.gmatch(message, '(.*)(%^%*)(.-)$') do
        if v=='nil' then v=nil end;
        tinsert(tpl, v)
    end
    if #tpl==0 then
        if message=='nil' then message=nil end;
        tinsert(tpl, message)
    end
    
    return tpl;
end

local MonitorMessagePriorities = {
    ["PRIORITY_HIGH"]       = "ALERT",
    ["PRIORITY_NORMAL"]     = "NORMAL",
    ["PRIORITY_LOW"]        = "BULK"
}
function LootMasterML:SendMonitorMessage(...)
    
    local numArgs = select("#", ...)
    
    local prio = nil
    if numArgs>0 then
        prio = select(numArgs, ...)
        prio = MonitorMessagePriorities[prio];
    end
    
    if prio then
        numArgs = numArgs - 1
    else
        prio = "NORMAL"
    end
    
    local out = ""
	for i=1,numArgs do
		if i > 1 then
			out = out .. "^*"
		end
		out = out .. tostring(select(i, ...))
	end
    
    local num = GetNumRaidMembers()
    if num>0 then
        -- we're in raid
        self:SendCommMessage("EPGPLootMasterML", format("MONITOR:%s", out), "RAID", nil, prio)
    else
        num = GetNumPartyMembers()
        if num>0 then
            --we're in party
            self:SendCommMessage("EPGPLootMasterML", format("MONITOR:%s", out), "PARTY", nil, prio)
        else
            --we're not grouped, send message to self for debugging purposes.
            self:SendCommMessage("EPGPLootMasterML", format("MONITOR:%s", out), "WHISPER", UnitName('player'), nil, prio)
        end
    end
end

--[[
	Event gets triggered when ML receives a message from a candidate
]]
function LootMasterML:CommandReceived(prefix, message, distribution, sender)

	local _,_,command, message = string.find(message, "^([%a_]-):(.*)$")
	command = strupper(command or '');
	message = message or '';
    
    local sendMonitorUpdate = (distribution == 'WHISPER')
    local preventMonitorUpdate = not sendMonitorUpdate;
	
	if command == 'WANT' then
        
		-- A candidate just told us he wants the item, lets just give it to him for testing purposes
        local _,_, itemID, response, enchantingSkill = string.find(message, "^([^\^]-)\^([^\^]-)\^(.*)$");
        enchantingSkill = tonumber(enchantingSkill) or 0;
        response = tonumber(response)
        
        if response == LootMaster.RESPONSE.PASS and enchantingSkill~=0 then
            response = LootMaster.RESPONSE.DISENCHANT;
        end        
        
        if self:GetLootID( itemID ) then        
            self:SetCandidateResponse( itemID, sender, response, preventMonitorUpdate );
            self:SetCandidateData( itemID, sender, 'enchantingSkill', enchantingSkill, preventMonitorUpdate );
            self:ReloadMLTableForLoot( itemID )
        else
            self:Debug( format("%s responded WANT for %s but loot or candidate not found (Already looted?)", sender or 'nil', itemID or 'nil'))
        end
		
	elseif command == 'GEAR' then

		-- A candidate just sent us his current gear. update the cache
        local _,_,itemID, iVersion, gear = string.find(message, "^([^\^]-)\^([^\^]-)\^(.*)$");
        local item1, item2 = strsplit('$', gear)
        
        if self:GetLootID( itemID ) then        
            self:SetCandidateData( itemID, sender, 'currentitem', item1, preventMonitorUpdate );
            self:SetCandidateData( itemID, sender, 'currentitem2', item2, preventMonitorUpdate );
            self:SetCandidateData( itemID, sender, 'version', iVersion, preventMonitorUpdate );
            self:SetCandidateResponse( itemID, sender, LootMaster.RESPONSE.WAIT, preventMonitorUpdate );
            self:ReloadMLTableForLoot( itemID )
            
            if sendMonitorUpdate then
                self:SendMonitorMessage( 'CANDIDATEGEAR', itemID, sender, item1, item2 )
            end
        else
            self:Debug( format("%s responded GEAR for %s but loot or candidate not found (Already looted?)", sender or 'nil', itemID or 'nil'))
        end
        
    elseif command == 'MONITOR' then
        
        -- Only handle monitor messages if we enabled them in the options screen.
        if not LootMaster.db.profile.monitor then return end;
        
        -- ignore our own monitor messages!
        if sender == UnitName('player') then return end;
        
        local monArgs = self:ParseMonitorMessage(message)
        local monCmd = tremove(monArgs, 1);
        
        if monCmd == 'ADDLOOT' then
            
            local itemLink, itemName, itemID, gpvalue, ilevel, itemBind, itemRarity, itemTexture, itemEquipLoc, gpvalue2 = unpack(monArgs)
            
            if not self.lootTable then self.lootTable={} end;
            
            if self.lootTable[itemID] then
                -- Is someone tinkering? the loot already exists. Return
                return;
            end
            
            self.lootTable[itemID] = {
                ['link']	        = itemLink,
                ['name']	        = itemName,
                
                ['lootmaster']      = sender,
                
                ['announced']       = true,
                ['mayDistribute']   = false,        
                
                ['id']              = itemID,
                ['itemID']          = itemID,
                ['itemid']          = itemID,
        
                ['gpvalue']	        = gpvalue or 0,
                ['gpvalue2']	    = gpvalue2 or 0,
                ['gpvalue_manual']	= gpvalue or 0,	
                ['ilevel']	        = ilevel or 0,
                ['binding']	        = itemBind,
                ['quality']         = itemRarity or 0,
                
                ['texture']         = itemTexture or '',
                
                ['equipLoc']	    = itemEquipLoc or '',
                
                ['started']	        = nil,
        
                ['rowdata']	        = {},
                ['candidates']	    = {}
            }
            
            self:ReloadMLTableForLoot( itemID )
            
        elseif monCmd == 'ADDCANDIDATE' then
            
            local itemID, candidate, roll = unpack(monArgs)
            local loot = self:GetLoot(itemID);
            
            if not loot or loot.mayDistribute or loot.lootmaster~=sender then
                -- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
                return;
            end
            
            self:SetCandidateData( itemID, candidate, 'roll', tonumber(roll) )
            
            self:ReloadMLTableForLoot( itemID )
            
        elseif monCmd == 'CANDIDATELIST' then
            
            -- in an attempt to speed the monitors up, lets just receive all candidate data for new loot in big chunk
            -- not candidate by candidate, updates hereafter are still one by one.
            
            local itemID, candidates = unpack(monArgs)
            local loot = self:GetLoot(itemID);
            
            if not loot or loot.mayDistribute or loot.lootmaster~=sender then
                -- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
                return;
            end
            
            if candidates then
                local cdata = {strsplit(' ', candidates)}
                
                for i=1, #cdata, 3 do
                    self:SetCandidateData( itemID, cdata[i], 'roll', tonumber(cdata[i+1]) or 0 )
                    
                    local response = tonumber(cdata[i+2]) or 0
                    local curResponse = tonumber(self:GetCandidateData( itemID, cdata[i], 'response' )) or 0
                    if response>curResponse then
                        self:SetCandidateData( itemID, cdata[i], 'response', response )
                    end
                end
            end
            
            self:ReloadMLTableForLoot( itemID )
            
        elseif monCmd == 'SETCANDIDATEDATA' then
            
            local itemID, candidate, name, value = unpack(monArgs)            
            local loot = self:GetLoot(itemID);
            
            if not loot or loot.mayDistribute or loot.lootmaster~=sender  then
                -- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
                return;
            end
            
            if not self:IsCandidate( itemID, candidate ) then return end;
            self:SetCandidateData( itemID, candidate, name, value )
            
            self:ReloadMLTableForLoot( itemID )
            
        elseif monCmd == 'CANDIDATEGEAR' then
            
            local itemID, candidate, item1, item2 = unpack(monArgs)            
            local loot = self:GetLoot(itemID);
            
            if not loot or loot.mayDistribute or loot.lootmaster~=sender  then
                -- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
                return;
            end
            
            if not self:IsCandidate( itemID, candidate ) then return end;
            
            if item1 then
                self:SetCandidateData( itemID, sender, 'currentitem', item1 );
            end
            if item2 then
                self:SetCandidateData( itemID, sender, 'currentitem2', item2 );
            end
            
            self:ReloadMLTableForLoot( itemID )
            
        elseif monCmd == 'REMOVELOOT' then
            
            local itemID = unpack(monArgs)
            local loot = self:GetLoot(itemID);
            
            if not loot or loot.mayDistribute or loot.lootmaster~=sender then
                -- Is someone tinkering? the loot doesnt exist or player is lootmaster. Return
                return;
            end
            
            self:RemoveLoot(itemID);
            self:UpdateUI();
            
        end
        
    else
		self:Print( format("MLRCV(%s): %s", tostring(command), tostring(message) ) )
	end
end

local deformat_cache = {}
function LootMasterML:Deformat(str, format)
  local pat = deformat_cache[format]
  if not pat then
    -- Escape special characters
    pat = format:gsub("[%^%$%(%)%%%.%[%]%*%+%-%?]",
                          function(c) return "%"..c end)
    pat = pat:gsub("%%%%([sd])", {
                             ["s"] = "(.-)",
                             ["d"] = "(%d+)",
                           })
    deformat_cache[format] = pat
  end
  return str:match(pat)
end

--[[
	Send a request to the specified player if he/she is interested in the given loot
]]
function LootMasterML:AskCandidateIfNeeded( link, candidate )

	local loot = self:GetLoot( link );
	if not loot then
		return self:Print( format("Could not ask player if needed because %s is not cached", link) )
	end
    
    -- Hardcoded time in secs until autopass
    self.timeout = LootMaster.db.profile.loot_timeout;

	self:SendCommand( "DO_YOU_WANT", format('%s^%s^%s^%s^%s^%s^%s^%s^%s^%s', loot.id or '', loot.gpvalue or -1, loot.ilevel or -1, loot.gpvalue2 or -1, loot.binding or '', loot.equipLoc or 0, loot.quality or 0, self.timeout or 60, loot.link, loot.texture or ''), candidate )

    self:SetCandidateResponse( link, candidate, LootMaster.RESPONSE.INIT );

	-- Update the ui
	self:ReloadMLTableForLoot( loot.id )
end

--[[
	Add loot to masterloot cache. This is where candidate responses are stored, plus the rows for the scrollingtable
]]
function LootMasterML:AddLoot( link, mayDistribute )
	if not link then return end;
	if not self.lootTable then self.lootTable={} end;
    
    if self.lootTable[link] then return link end;

	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(link)
    local _,_,itemID = strfind(itemLink, 'Hitem:(%d+)');
    if not itemID or not itemName then return end;

	if self.lootTable[itemID] then return itemID end;

	-- Calc the EPGP values for this item
	local gpvalue, gpvalue2, ilevel = GetGPValue( itemLink );

	-- See if the item is BoP, BoE or BoU
	local itemBind = LootMaster:GetItemBinding( itemLink )	

	self.lootTable[itemID] = {
		['link']	        = itemLink,
		['name']	        = itemName,
        
        ['announced']       = true,
        ['mayDistribute']   = mayDistribute,        
        
        ['id']              = itemID,
        ['itemID']          = itemID,
        ['itemid']          = itemID,

		['gpvalue']	        = gpvalue or 0,
        ['gpvalue2']	    = gpvalue2 or 0,
        ['gpvalue_manual']	= gpvalue or 0,	
		['ilevel']	        = ilevel or 0,
		['binding']	        = itemBind,
        ['quality']         = itemRarity or 0,
        
        ['texture']         = itemTexture or '',
		
		['equipLoc']	    = itemEquipLoc or '',
		
		['started']	        = nil,

		['rowdata']	        = {},
		['candidates']	    = {}
	}
    
    -- Are we lootmaster for this loot? Lets send out a monitor message about the added loot
    if self.lootTable[itemID].mayDistribute then
        self:SendMonitorMessage( 'ADDLOOT', itemLink, itemName, itemID, gpvalue or 0, ilevel or 0, itemBind, itemRarity or 0, itemTexture or '', itemEquipLoc or '', gpvalue2 or 0, "PRIORITY_HIGH" )
    end

	return itemID;
end

function LootMasterML:AnnounceLoot( loot )
    loot = self:GetLoot( loot )
    if not loot then return end;
    
    if loot.announced == true then return end;
    
    for candidate, index in pairs(loot.candidates) do
        if self:GetCandidateData( loot.id, candidate, 'response') == LootMaster.RESPONSE.NOTANNOUNCED then
            self:AskCandidateIfNeeded( loot.id, candidate )
        end
    end
    loot.announced = true;
end


--[[
	Get the lootdata from masterloot cache. This is where candidate responses are stored, plus the rows for the scrollingtable. Returns nil if not found
]]
function LootMasterML:GetLoot( link )
	if not link then self:Debug('getloot: !link'); return nil end;
	if not self.lootTable then self:Debug('getloot: !lootTable'); return nil end;	
    
    if self.lootTable[link] then return self.lootTable[link] end;
    
    local _,_,itemID = strfind(link, 'Hitem:(%d+)');
    if not itemID then self:Debug('getloot: !itemID: ' .. tostring(itemID)); return end;
    if not self.lootTable[itemID] then self:Debug('getloot: !lootTable[itemID]: ' .. tostring(itemID)); return end;

	return self.lootTable[itemID];
end

--[[
	Get the lootdata from masterloot cache. This is where candidate responses are stored, plus the rows for the scrollingtable. Returns nil if not found
]]
function LootMasterML:GetLootID( link )
	if not link then self:Debug('getlootid: !link'); return nil end;
	if not self.lootTable then self:Debug('getlootid: !lootTable'); return nil end;	
    
    if self.lootTable[link] then return link end;
    
    local _,_,itemID = strfind(link, 'Hitem:(%d+)');
    if not itemID then self:Debug('getlootid: !itemID: ' .. tostring(itemID)); return end;
    if not self.lootTable[itemID] then self:Debug('getloot: !lootTable[itemID]: ' .. tostring(itemID)); return end;

	return itemID;
end

--[[
	Removes the given loot from the masterloot cache
]]
function LootMasterML:RemoveLoot( link )
	if not link then return end;
	if not self.lootTable then return end;	
    
    if self.lootTable[link] then
        
        -- Are we lootmaster for this loot? Lets send out a monitor message about the removed loot
        if self.lootTable[link].mayDistribute then
            self:SendMonitorMessage( 'REMOVELOOT', link, "PRIORITY_HIGH" )
        end
    
        self.lootTable[link] = nil;
        if self.frame and self.frame.currentLoot and self.frame.currentLoot.itemID == link then
            self.frame.currentLoot = nil;
        end
        return true
    end
    
    local _,_,itemID = strfind(link, 'Hitem:(%d+)');
    if not itemID or not self.lootTable[itemID] then return self:Debug(format('RemoveLoot: not found %s %s(%s)', link or 'nil', itemID or 'nil', type(itemID))) end;
    
    if self.frame and self.frame.currentLoot and self.frame.currentLoot.itemID == itemID then
        self.frame.currentLoot = nil;
    end

    self:Debug('RemoveLoot: '..link)
    
    -- Are we lootmaster for this loot? Lets send out a monitor message about the removed loot
    if self.lootTable[itemID].mayDistribute then
        self:SendMonitorMessage( 'REMOVELOOT', itemID, "PRIORITY_HIGH" )
    end

	self.lootTable[itemID] = nil;
	return true;
end

function LootMasterML:GetEPGP( player )
    if not EPGP or not EPGP.GetEPGP then return nil, nil, nil, nil end;
    local ret, ep, gp, alt = pcall(EPGP.GetEPGP, EPGP, player)
    if not ret then return nil, nil, nil, nil end;
    
    local ret, minEP = pcall(EPGP.GetMinEP, EPGP)
    if not ret or not minEP then minEP=0 end;
    
    return ep, gp, alt, minEP;
end

function LootMasterML:GetEP( player )
    local ep, gp = self:GetEPGP( player );
    if not ep then return -1 end;
    return ep or 0
end

function LootMasterML:GetMinEPMatch( player )
    local ep, gp, alt, minEP = self:GetEPGP( player );
    if not ep or not minEP then return 'n' end;
    
    if ep>=minEP then return 'y' end
    return 'n';
end

function LootMasterML:GetGP( player )
    local ep, gp = self:GetEPGP( player );
    if not gp then return -1 end;
    return gp or 1
end

function LootMasterML:GetPR( player )
    local ep, gp = self:GetEPGP( player );
    if not gp or not ep then return -1 end;
    if gp>0 then return ep/gp end;
    return 0;
end

--[[
	Adds a candidate (player eligible to receive loot) to the masterloot cache
	for the given loot
]]
function LootMasterML:AddCandidate( loot, candidate )
	local itemID = self:AddLoot( loot )
	if not itemID then return end
	if not candidate then return end
	if not self.lootTable[itemID] then return end

	-- Just return the itemname if the candidate already exists
	if self.lootTable[itemID]['candidates'][candidate] then return itemID end;
    
    math.randomseed(math.random(0,2147483646)+(GetTime()*1000));
    local randomRoll = math.random(1, 99) + math.random(1, 10000)/10000;
    
    local guildRankName = ''
    if self.guildInfo and self.guildInfo[candidate] then
        guildRankName = self.guildInfo[candidate]
    end

	tinsert( self.lootTable[itemID]['rowdata'],
		 {  ["onclick"]         = function(arg1, button, down) if button=='RightButton' then self:OnCandidateRowRightClick(candidate, loot, arg1) end end,
			["cols"] = {
				{["name"]       = 'name',
				 ["value"]      = candidate,
                 ["onenter"]    = addon.ShowCandidateCellPopup,
                 ["onleave"]    = addon.HideGearCellPopup,
                 ["color"]      = self.GetCandidateCellColor,
                 ["colorargs"]  = { self, candidate, itemID},
                 ["onenterargs"]= { self, candidate, itemID },
                 ["onleaveargs"]= { self, candidate, itemID }
                },
                 
                {["name"]       = 'guildrank',
				 ["value"]      = guildRankName,
                 ["color"]      = self.GetCandidateCellColor,
                 ["colorargs"]  = {self, candidate, itemID}},  
                 
                {["name"]       = 'response',
				 ["value"]      = LootMaster.RESPONSE.NOTANNOUNCED,
                 ["args"]       = {self, candidate, itemID},
                 ["userDraw"]   = addon.SetCandidateResponseCellUserDraw,
                 ["color"]      = self.GetCandidateCellColor,
                 ["colorargs"]  = {self, candidate, itemID}},
				
				{["name"]       = 'EP',
				 ["value"]      = self.GetEP,
                 ["userDraw"]   = addon.SetCellEPGPNumberFormatted,
                 ["args"]       = {self, candidate, self.GetEP}},
                 
                {["name"]       = 'GP',
				 ["value"]      = self.GetGP,
                 ["userDraw"]   = addon.SetCellEPGPNumberFormatted,
                 ["args"]       = {self, candidate, self.GetGP}},
                 
                {["name"]       = 'PR',
				 ["value"]      = self.GetPR,
                 ["userDraw"]   = addon.SetCellEPGPNumberFormatted,
                 ["args"]       = {self, candidate, self.GetPR}},
                 
                {["name"]       = 'roll',
				 ["value"]      = randomRoll,
                 ["userDraw"]   = addon.SetCandidateRollCellUserDraw,                 
                 ["onenter"]    = addon.ShowRollCellPopup,
                 ["onleave"]    = addon.HideGearCellPopup,
                 ["onenterargs"]= { self,candidate,itemID },
                 ["onleaveargs"]= { self,candidate,itemID },
                 ["args"]       = {self, candidate, itemID}},
                 
                {["name"]       = ' ',
				 ["value"]      = self.GetMinEPMatch,
                 ["userDraw"]   = self.EmptyCellOwnerDraw,
                 ["args"]       = {self, candidate}},     -- spacer, actually used for sorting the minEP matches first.

				{["name"]       = 'currentilevel',
				 ["value"]      = '',
                 ["args"]       = {self, candidate, itemID},
                 ["userDraw"]   = addon.SetGearCelliLVL},

				{["name"]       = 'currentgp',
				 ["value"]      = '',
                 ["args"]       = {self, candidate, itemID},
                 ["userDraw"]   = addon.SetGearCellGP},

				{["name"]       = 'currentitem',
				 ["value"]      = '',
                 ["userDraw"]   = addon.SetGearCellOwnerDraw,
                 ["onenter"]    = addon.ShowGearCellPopup,
                 ["onleave"]    = addon.HideGearCellPopup,
                 ["onclick"]    = addon.OnGearCellClick,
                 ["onenterargs"]= { self,candidate,itemID,'currentitem' },
                 ["onleaveargs"]= { self,candidate,itemID,'currentitem' },
                 ["onclickargs"]= { self,candidate,itemID,'currentitem' }
                },
                
                {["name"]       = 'currentitem2',
				 ["value"]      = '',
                 ["userDraw"]   = addon.SetGearCellOwnerDraw,
                 ["onenter"]    = addon.ShowGearCellPopup,
                 ["onleave"]    = addon.HideGearCellPopup,
                 ["onclick"]    = addon.OnGearCellClick,
                 ["onenterargs"]= { self,candidate,itemID,'currentitem2' },
                 ["onleaveargs"]= { self,candidate,itemID,'currentitem2' },
                 ["onclickargs"]= { self,candidate,itemID,'currentitem2' }
                },
                
                {["value"]      = ' '},     -- spacer
			}
		 })
	
	self.lootTable[itemID]['candidates'][candidate] = #(self.lootTable[itemID]['rowdata']);
    
    -- Are we lootmaster for this loot? Lets send out a monitor message about the added candidate
    if self.lootTable[itemID].mayDistribute and self.lootTable[itemID].candidatesSent then
        self:SendMonitorMessage( 'ADDCANDIDATE', itemID, candidate, randomRoll )
    end

	return itemID;	
end

function LootMasterML:SendCandidateListToMonitors( itemID )
    local loot = self:GetLoot( itemID )
    if not loot then return nil end;
        
    local candidata = {}
    
    for candidate, cIndex in pairs(loot.candidates) do
        local roll = self:GetCandidateData( itemID, candidate, 'roll' ) or 0
        local response = self:GetCandidateData( itemID, candidate, 'response' ) or LootMaster.RESPONSE.NOTANNOUNCED
        tinsert( candidata, candidate );
        tinsert( candidata, roll );
        tinsert( candidata, response );
    end
    
    self.lootTable[itemID].candidatesSent = true;
    
    self:SendMonitorMessage( 'CANDIDATELIST', itemID, strjoin(' ', unpack(candidata)) ) --, 'PRIORITY_HIGH' )    
end

--[[
	Returns true if the candidate already has been cached for the given loot
]]
function LootMasterML:IsCandidate( loot, candidate )
	
    if not loot then return false end;
    
    loot = self:GetLoot(loot);    
    if not loot or not loot.id then return false end;

	if not loot
		or not self.lootTable
		or not self.lootTable[loot.id]
		or not self.lootTable[loot.id]['candidates'] then return false end;

	local rowID = self.lootTable[loot.id]['candidates'][candidate];
	if not rowID then return false end;

	return true;
end

--[[
	Stores some data in the candidate cache for the given loot
]]
function LootMasterML:SetCandidateData( loot, candidate, name, value, preventMonitorUpdate )

	local itemID = self:AddCandidate( loot, candidate );
	if not itemID then return end;

	local rowID = self.lootTable[itemID]['candidates'][candidate];
	if not rowID then return end;

	for index,data in pairs( self.lootTable[itemID]['rowdata'][rowID]['cols'] ) do
		if data and data['name'] == name then
			self.lootTable[itemID]['rowdata'][rowID]['cols'][index]['value'] = value;
            
            -- Are we lootmaster for this loot? Lets send out a monitor message about the data change
            if self.lootTable[itemID].mayDistribute then
                self:SendMonitorMessage( 'SETCANDIDATEDATA', itemID, candidate, name, value )
            end
          
			return value
		end
	end

	self.lootTable[itemID]['rowdata'][rowID][name] = value;    
    
    -- Are we lootmaster for this loot? Lets send out a monitor message about the data change
    if self.lootTable[itemID].mayDistribute and not preventMonitorUpdate then
        
        -- Default priority
        local prio = "PRIORITY_NORMAL"
        
        -- High priority on response changes
        if name == 'response' then
            prio = "PRIORITY_HIGH"
        end
        
        -- Dont need the following changes
        if name=='roll' or name=='version' or name=='currentitem' or name=='currentitem2' then
            prio = nil
        end
        
        -- Do not send status updates before candidate list has been sent.
        if not self.lootTable[itemID].candidatesSent and name=='response' then
            prio = nil
        end
        
        if prio then
            self:SendMonitorMessage( 'SETCANDIDATEDATA', itemID, candidate, name, value, prio )
        end
    end
    
	return value;
end

function LootMasterML:SetCandidateResponse( link, candidate, response, preventMonitorUpdate )    
    return self:SetCandidateData( link, candidate, 'response', response, preventMonitorUpdate )
end

--[[
	Retrieves some data in the candidate cache for the given loot
]]
function LootMasterML:GetCandidateData( loot, candidate, name )

    if not loot then return end;
	loot = self:GetLoot(loot)

	if not loot or not loot.id
		or not self.lootTable
		or not self.lootTable[loot.id]
		or not self.lootTable[loot.id]['candidates'] then return end;

	local rowID = self.lootTable[loot.id]['candidates'][candidate];
	if not rowID then return end;

	for index,value in pairs( self.lootTable[loot.id]['rowdata'][rowID]['cols'] ) do
		if value and value['name'] == name then
			return self.lootTable[loot.id]['rowdata'][rowID]['cols'][index]['value']
		end
	end

	return self.lootTable[loot.id]['rowdata'][rowID][name];
end

--[[
	Tries to give the loot to the given candidate
]]
function LootMasterML:GiveLootToCandidate( link, candidate, lootType, gp )

	local candidateID = nil;
	local slotID = nil;
	local loot = self:GetLoot(link)
    
    if not loot or not loot.id then
        return self:Print( format("Could not send %s to %s, loot not found in cache", tostring(link), tostring(candidate) ) )
    end

	-- Look for the candidateID
	for cID = 1, 40 do
		local cName = GetMasterLootCandidate(cID)
		if cName and cName==candidate then
			candidateID = cID;
			break
		end;
	end;

	-- Look for the lootslotID
	for sID = 1, GetNumLootItems() do
		local _, sName = GetLootSlotInfo(sID);
		if sName and sName==loot.name then
			slotID = sID;
			break
		end;
	end;

	if not slotID then
		return self:Print( format("Could not send %s to %s, lootslotID not found (already looted or lootwindow closed?) ", tostring(loot.link), tostring(candidate) ) )
	end
	if not candidateID then
		return self:Print( format("Could not send %s to %s, candidate not found (offline, left group?)", tostring(loot.link), tostring(candidate) ) )
	end
    
    -- [[ lootType == self.LOOTTYPE.BANK, self.LOOTTYPE.DISENCHANT, self.LOOTTYPE.GP ]]--
    -- [[ gp = amount of gp awarded to this item ]] --
    
    if lootType ~= LootMaster.LOOTTYPE.GP then
        gp = 0;
    end
    
    self:SetCandidateData( loot.id, candidate, 'lootType', lootType or 0 );
    self:SetCandidateData( loot.id, candidate, 'lootGP', tonumber(gp) or 0 );

	GiveMasterLoot( slotID, candidateID )
	
end

function LootMasterML:CacheGuildInfo( event )    
    local num = GetNumGuildMembers(false);
    for i=1, num do repeat
        local name, rank = GetGuildRosterInfo(i)
        if name and rank then
            self.guildInfo[name] = rank;
        end
    until true end    
end

local LootMessages = {
        ["SINGLE"]      = "([^%s]+) receives loot: (.+)\.",
	    ["SINGLESELF"]  = "You receive loot: (.+)\.",
	    ["MULTI"]       = "([^%s]+) receives loot: (.-)x(%d+)\.",
	    ["MULTISELF"]   = "You receive loot: (.-)x(%d+)\."
};

--[[ Someone received loot, lets see if it's in our cache. Send all candidates 
     a message about this, so they can update their UI and register the lootdrop
     in a raidtracker for example ]]--
function LootMasterML:CHAT_MSG_LOOT( event, message )
    
    local iStart, iEnd, sPlayer, sLink, iCount;    
    
    self:Debug("CHAT_MSG_LOOT: " .. message)
		
	if(string.find(message, LootMessages.MULTI)) then
			iStart, iEnd, sPlayer, sLink, iCount = string.find(message, LootMessages.MULTI);
            
	elseif(string.find(message, LootMessages.MULTISELF)) then
			iStart, iEnd, sLink, iCount = string.find(message, LootMessages.MULTISELF);
			sPlayer = UnitName("player");
            
    elseif(string.find(message, LootMessages.SINGLE)) then
			iStart, iEnd, sPlayer, sLink = string.find(message, LootMessages.SINGLE);
			iCount = 1;
            
	elseif(string.find(message, LootMessages.SINGLESELF)) then
			iStart, iEnd, sLink = string.find(message, LootMessages.SINGLESELF);
			iCount = 1;
			sPlayer = UnitName("player");
	end
    
    if not sPlayer or not sLink or not iCount or tonumber(iCount)>1 then
        self:Debug("!player or !link or !iCount or iCount>1");
        return
    end;
    
    self:Debug(format("%s: %sx%s", sPlayer or 'nil', sLink or 'nil', iCount or 0 ))
    
    -- ok, someone looted something... lets see if we can find it in our cache    
    local loot = self:GetLoot( sLink )
    
    -- Did we find anything?
    if not loot or not loot.candidates then return self:Debug('loot not found') end;
    
    -- Did we own this loot or are we just monitoring it?
    if not loot.mayDistribute then return self:Debug('only monitor loot, return') end;
    
    -- We did, now get some info about the loot adressee
    local lootType = self:GetCandidateData( sLink, sPlayer, 'lootType' ) or LootMaster.LOOTTYPE.UNKNOWN;
    local lootGP = tonumber(self:GetCandidateData( sLink, sPlayer, 'lootGP' )) or -1;
    
    if lootType==LootMaster.LOOTTYPE.GP and lootGP~=0 then
        if CanEditOfficerNote() and EPGP and EPGP.IncGPBy then
            EPGP.IncGPBy(EPGP, sPlayer, sLink or 'nil', lootGP)
        else
            self:Print( format("Could not increase GP in officernotes for %s %s", sPlayer or 'unknown player', sLink or 'unknown loot') )
        end
    end;
    
    -- now send everyone in raid/party/candidates some info about the drop so they can update their ui
    if GetNumRaidMembers()>0 then
        self:Debug("send to raid");
        self:SendCommand( 'LOOTED', format('%s^%s^%s^%s', sPlayer, sLink, lootType, lootGP ), 'RAID' );
    elseif GetNumPartyMembers()>0 then
        self:Debug("send to party");
        self:SendCommand( 'LOOTED', format('%s^%s^%s^%s', sPlayer, sLink, lootType, lootGP ), 'PARTY' );
    else
        self:Debug("send to candidates");
        for candidate, id in pairs(loot.candidates) do
            self:SendCommand( 'LOOTED', format('%s^%s^%s^%s', sPlayer, sLink, lootType, lootGP ), candidate );
        end
    end
    
    -- Since it's looted, remove it from the cache and update the UI
    self:RemoveLoot(sLink);
    self:UpdateUI();
    
    self:Debug("CHAT_MSG_LOOT end");
    
end

--[[
	Event triggers when the master looter opens the popup on the loot screen
	In here we try to see what loot got selected and find the candidates for the
	loot. Ask the candidates if they want the loot. Tho, cache everything and make sure
	we don't ask the candidates more than one time (this could happen when the lootwindow
	gets opened more than one time)
]]
function LootMasterML:OPEN_MASTER_LOOT_LIST()

	-- Close the default confirm window
	StaticPopup_Hide("CONFIRM_LOOT_DISTRIBUTION");
	
	--[[
	-- Some values we probably need
	LootFrame.selectedLootButton
	LootFrame.selectedSlot
	LootFrame.selectedQuality
	LootFrame.selectedItemName
	]]

	-- local lootIcon, lootName, lootQuantity, rarity = GetLootSlotInfo(LootFrame.selectedSlot);
	local _, lootName, lootQuantity, rarity = GetLootSlotInfo(LootFrame.selectedSlot);
	local link = GetLootSlotLink(LootFrame.selectedSlot)

	-- A little sanity check; lets see if quantity == 1
	if lootQuantity~=1 then
		self:Print( format("Could not redistribute %s because quantity != 1 (%s). Please handle it manually. Create a ticket on curseforge if this happens often.", link, lootQuantity) )
		return;
	end

	-- itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture
	-- local _,_,_, itemLevel, itemMinLevel, itemType, itemSubType,_, itemEquipLoc, itemTexture = GetItemInfo(link)

	-- Register the loot in the loottable
	local lootID = self:AddLoot(link, true);
    if not lootID then return self:Print('Could not register loot') end;
    
    -- Lootmaster is handling the loot, so lets close the default popup
    CloseDropDownMenus();
    
    -- Auto announce?
    local autoAnnounce = rarity >= (LootMaster.db.profile.auto_announce_threshold or 4);    
    if LootMaster.db.profile.auto_announce_threshold == 0 then autoAnnounce=false end;

	-- Ok Lets see. Who are the candidates for this slot?
	for candidateID = 1, 40 do repeat
		local candidate = GetMasterLootCandidate(candidateID)
		if not candidate then break end;

		if not self:IsCandidate( lootID, candidate ) then

			-- Create the candidate for link;
			self:AddCandidate( lootID, candidate );
		end	
	until true end
    
    -- Set the loot status to not announced.
    self.lootTable[lootID].announced = false;
    
    -- See if we have to auto announce
    if autoAnnounce then
        self:AnnounceLoot( lootID )
    end
    
    self:ReloadMLTableForLoot( lootID )
    
    self:SendCandidateListToMonitors( lootID )
    
    if (not autoAnnounce) and LootMaster.db.profile.auto_announce_threshold ~= 0 and LootMaster.db.profile.UseAutoLoot and LootMaster.db.profile.AutoLooter~='' then
        -- loot is below autoAnnounce threshold and UseAutoLoot is enabled.
        -- Lets see if we have a BoE or BoU item.
        
        local binding = self.lootTable[lootID].binding;
        if not binding or binding=='use' or binding=='equip' then
            -- We have an item that doesn't bind or is BoE or BoE, try to give the loot.
            
            if self:IsCandidate( lootID, LootMaster.db.profile.AutoLooter or '' ) then 
                self:Print(format('Auto looting %s to %s', link or 'nil', LootMaster.db.profile.AutoLooter or 'nil'))
                self:GiveLootToCandidate( lootID, LootMaster.db.profile.AutoLooter or '', LootMaster.LOOTTYPE.BANK)
            else
                self:Print(format('Auto looting of %s to %s failed. Not a candidate for this loot.', link or 'nil', LootMaster.db.profile.AutoLooter or 'nil'))
            end
        end
    end
    
end